/* Engine */
#include<stdio.h>
#include<stdlib.h>
#include "engine.h"

struct Game game;
static int dir_x[4]={-1,0,1,0},dir_y[4]={0,-1,0,1};

void new_game()
{
	game.level=1;
	game.lives=3;
	game.score=0;
}

void new_level(int level)
{
	int i,j,x,y;

	srand(12345*level);
	for(i=0;i<MAPWIDTH*MAPHEIGHT;i++) game.map[i]=M_WALL;
	// Now create the path sections
	for(i=0;i<35;i++) {
		// Horizontal
		x=rand()%(MAPWIDTH-9);
		x&=~1;	// Only even columns
		x+=2;
		y=rand()%(MAPHEIGHT-3);
		y&=~1;
		y+=2;
		for(j=y*MAPWIDTH+x;j<y*MAPWIDTH+x+6;j++) {
			game.map[j]=M_OPENPATH;
		}
	}
	for(i=0;i<40;i++) {
		// Horizontal
		x=rand()%(MAPWIDTH-3);
		x&=~1;	// Only even columns
		x+=2;	// Margin on the left
		y=rand()%(MAPHEIGHT-9);
		y&=~1;
		y+=2;	// Margin on the top
		for(j=y*MAPWIDTH+x;j<(y+6)*MAPWIDTH+x;j+=MAPWIDTH) {
			game.map[j]=M_OPENPATH;
		}
	}
	// Now place the treasures (need 25 to finish level)
	int active=0;
	for(i=0;i<40 || active<25;i++) {
		x=rand()%(MAPWIDTH-5);
		x&=~1;
		x+=2;
		y=rand()%(MAPHEIGHT-5);
		y&=~1;
		y+=2;
		if(game.map[y*MAPWIDTH+x]!=M_TREASURE) active++;
		game.map[y*MAPWIDTH+x]=M_TREASURE;
	}
	game.player.x=TILEWIDTH*2;
	game.player.y=TILEHEIGHT*2;
	game.player.delay=0;
	game.player.dir=-1;
	game.treasures=25;
	game.npc[0].delay=0;
	game.npc[0].dir=-1;
	game.npc[0].x=-TILEWIDTH;
	game.npc[1].delay=0;
	game.npc[1].dir=-1;
	game.npc[1].x=-TILEWIDTH;
	if(((game.level/3)%2)==0) play_sound(S_RUSH);
	else play_sound(S_ALONE);
}

void draw_map()
{
	int x,y,i=0;
	for(y=0;y<MAPHEIGHT;y++) {
		for(x=0;x<MAPWIDTH;x++) {
			if(game.map[i]==M_TREASURE)
				draw_tile(M_OPENPATH,x*TILEWIDTH,y*TILEHEIGHT);
			draw_tile(game.map[i++],x*TILEWIDTH,y*TILEHEIGHT);
		}
	}
	draw_tile(M_PLAYER,game.player.x,game.player.y);
	draw_tile(M_NPC,game.npc[0].x,game.npc[0].y);
	draw_tile(M_NPC,game.npc[1].x,game.npc[1].y);
	for(i=0;i<3;i++) {
		draw_tile(M_LEVEL+i,TILEWIDTH*i,0);
		draw_tile(M_SCORE+i,TILEWIDTH*(i+5),0);
	}
	draw_tile(M_WHITE,TILEWIDTH*12,0);
	draw_tile(M_WHITE,TILEWIDTH*13,0);
	draw_tile(M_TREASURE,TILEWIDTH*13,0);
	draw_tile(M_WHITE,TILEWIDTH*16,0);
	draw_tile(M_WHITE,TILEWIDTH*17,0);
	draw_tile(M_PLAYER,TILEWIDTH*17,0);
	// Now decode the number into decimal.
	int num;
	num=game.level;
	draw_number((num/10)%10,TILEWIDTH*3,0);
	draw_number(num%10,TILEWIDTH*4,0);

	num=game.score;
	draw_number((num/1000)%10,TILEWIDTH*8,0);
	draw_number((num/100)%10,TILEWIDTH*9,0);
	draw_number((num/10)%10,TILEWIDTH*10,0);
	draw_number(num%10,TILEWIDTH*11,0);	

	num=game.treasures;
	if(num<0) num=0;
	draw_number((num/10)%10,TILEWIDTH*14,0);
	draw_number(num%10,TILEWIDTH*15,0);

	num=game.lives;
	if(num<0) num=0;
	draw_number(num%10,TILEWIDTH*18,0);
	
	if(game.lives<1) {
		for(i=0;i<5;i++) {
			draw_number(10+i,TILEWIDTH*(MAPWIDTH/2-2+i),TILEHEIGHT*5);
		}
	}
}

void move_character(int direction)
{
	if(game.lives<1) return;
	
//	fprintf(stderr,"move_character(%d::<=%d) [%d delay]\n",direction,game.player.dir,game.player.delay);
	if(game.player.delay!=0 && game.player.dir!=direction) {
		return;	// Not ready yet.
	}
	if(game.player.delay>0) return;	// Not ready yet.
	// Check if the target is a valid square
	int x=game.player.x;
	int y=game.player.y;
	if(game.player.delay>0) {
		x+=dir_x[game.player.dir]*game.player.delay;
		y+=dir_y[game.player.dir]*game.player.delay;
	}
	x/=TILEWIDTH;
	x+=dir_x[direction];
	y/=TILEHEIGHT;
	y+=dir_y[direction];
	
	int i=y*MAPWIDTH+x;
	if(game.map[i]==M_OPENPATH || game.map[i]==M_VISITEDPATH || game.map[i]==M_TREASURE) {
		// Start moving
		game.player.delay+=TILEWIDTH;
		game.player.dir=direction;
	}
	//fprintf(stderr,"move_character(%d<=%d) @ ->%d,%d [%d delay]\n",direction,game.player.dir,x,y,game.player.delay);
}

void explode_path()
{
	int x,y;
	
	if(game.lives<1) {
		new_game();
		new_level(game.level);
		return;
	}
	x=game.player.x;
	y=game.player.y;
	if(game.player.dir>=0 && game.player.delay) {
		x+=dir_x[game.player.dir]*game.player.delay;
		y+=dir_y[game.player.dir]*game.player.delay;
	}
	x/=TILEWIDTH;
	y/=TILEHEIGHT;
	game.map[y*MAPWIDTH+x]=M_VISITEDPATH;
	if(x>1)
		game.map[y*MAPWIDTH+x-1]=M_VISITEDPATH;
	if(x<MAPWIDTH-1)
		game.map[y*MAPWIDTH+x+1]=M_VISITEDPATH;
	if(y>1)
		game.map[(y-1)*MAPWIDTH+x]=M_VISITEDPATH;
	if(y<MAPHEIGHT-1)
		game.map[(y+1)*MAPWIDTH+x]=M_VISITEDPATH;
	play_sound(S_BALLBAT);
}

void next_frame_tick();

void next_frame(int elapsed)
{
	while(elapsed>5) {
		next_frame_tick();
		elapsed-=5;
	}
}

void next_frame_tick()
{
	if(game.lives<1) return;

	int i;
	// Now intersect the npcs with the player.
	int caught=0;
	for(i=0;i<2;i++) {
		if(game.player.x==game.npc[i].x) {
			if(game.player.y<game.npc[i].y) {
				if(game.player.y+TILEHEIGHT>=game.npc[i].y) {
					caught=1;
				}
			} else {
				if(game.player.y<=game.npc[i].y+TILEHEIGHT) {
					caught=1;
				}
			}
		}
		if(game.player.y==game.npc[i].y) {
			if(game.player.x<game.npc[i].x) {
				if(game.player.x+TILEWIDTH>=game.npc[i].x) {
					caught=1;
				}
			} else {
				if(game.player.x<=game.npc[i].x+TILEWIDTH) {
					caught=1;
				}
			}
		}
	}
	if(caught) {
		play_sound(S_BOMB);
		game.lives--;
		new_level(game.level);	// Same level reset.
	}
//	if((game.player.x%TILEWIDTH)==0 && (game.player.y%TILEHEIGHT)==0) {
	if((game.player.delay%16)==0) {
		int i=game.player.x/16+game.player.y/16*MAPWIDTH;
		if(game.map[i]==M_TREASURE) {
			game.treasures--;
			game.score++;
			play_sound(S_POP);
		}
		game.map[i]=M_VISITEDPATH;
	}

	// Now movement time
	if(game.player.dir>=0 && game.player.delay) {
		game.player.x+=dir_x[game.player.dir];
		game.player.y+=dir_y[game.player.dir];
	}
	if(game.player.delay>0)
		game.player.delay--;
	if(game.player.delay==0) game.player.dir=D_NONE;
	
	if(game.treasures<=0) {
		// finish level.
		game.level++;
		play_sound(S_APPLAUSE);
		new_level(game.level);
	}
	int headstart=(game.level>=8)?2:(10-game.level);
	if(25-game.treasures<headstart) {
		game.npc[0].x=-TILEWIDTH;
		game.npc[1].x=-TILEWIDTH;
		return;
	}
	int j;
	static int handicap=0;
	handicap=1-handicap;
	// Slow down the NPCs a bit.
	if(handicap) {
		return;
	}
	for(i=0;i<2;i++) {
		if(game.npc[i].delay>0) {
			game.npc[i].x+=dir_x[game.npc[i].dir];
			game.npc[i].y+=dir_y[game.npc[i].dir];
			game.npc[i].delay--;
			continue;
		}
		if(game.npc[i].x<0) {
			game.npc[i].x=2*TILEWIDTH;
			game.npc[i].y=2*TILEHEIGHT;
			game.npc[i].delay=0;
			game.npc[i].dir=D_LEFT;
		}
		// Now look around in all four directions to choose a chase direction
		//int npcdir=rand()%4;
		int npcdir=game.npc[i].dir;
		npcdir=(npcdir+1)%4;	// Default to continuing straight along

		// Otherwise follows the right hand rule to chase.		
		int x=game.npc[i].x/TILEWIDTH;
		int y=game.npc[i].y/TILEHEIGHT;
		int dvisited=-1;
		int dopen=-1;
		// Count choices (less go backwards), and pick one at random.
		int choiceVisited=0;
		int choiceOpen=0;
		for(j=0;j<4;j++) {
			if(npcdir!=(game.npc[i].dir+2)%4) {
				switch(game.map[(x+dir_x[npcdir])+(y+dir_y[npcdir])*MAPWIDTH]) {
				case M_OPENPATH:
				case M_TREASURE:
					choiceOpen++;
					break;
				case M_VISITEDPATH:
					choiceOpen++;
					break;
				}
			}
			npcdir=(npcdir+1)%4;
		}
		if(choiceVisited+choiceOpen==0) {
			game.npc[i].dir=(game.npc[i].dir+2)%4;
			game.npc[i].delay=16;
			continue;
		}
		// Now there is a valid set of choices:
		int pref=0;
		if(choiceVisited>0) pref=rand()%choiceVisited;
		else if(choiceOpen>0) pref=rand()%choiceOpen;
		for(j=0;j<4;j++) {
			if(npcdir!=(game.npc[i].dir+2)%4) {
				int m=game.map[(x+dir_x[npcdir])+(y+dir_y[npcdir])*MAPWIDTH];
				if(choiceVisited==0) {
					if(m==M_OPENPATH || m==M_TREASURE) {
						dopen=npcdir;
						pref--;
						if(pref==-1) break;
					}
					if(m==M_VISITEDPATH) {
						dvisited=npcdir;
					}
				} else {
					if(m==M_VISITEDPATH) {
						dvisited=npcdir;
						pref--;
						if(pref==-1) break;
					}
					if(m==M_OPENPATH || m==M_TREASURE) {
						dopen=npcdir;
					}
				}
			}
			npcdir=(npcdir+1)%4;
		}
		if(dvisited>-1) {
			game.npc[i].dir=dvisited;
			game.npc[i].delay=16;
		} else if(dopen>-1) {
			game.npc[i].dir=dopen;
			game.npc[i].delay=16;
		} else {
			game.npc[i].dir=D_NONE;	// Bug
			game.npc[i].delay=16;
		}
	}
}
